#!~/virtualenv-1.9/myEV/bin/python
# encoding=utf-8

""" Module for creating exercises using given ontology. """

from __future__ import unicode_literals
from language_tools import isVerb, phraseGenerator, randomNameGenerator, \
        randomPlaceGenerator, formWQuestion, sentenceFromPhrase, \
        isContextFree, declineNounPhrase
from rdflib import Graph, Literal, URIRef, BNode, Namespace
from rdflib.namespace import RDF, RDFS, FOAF
from ontology import findAll, findAllKeywords, EXERCISES, KEYWORD
import random
import json


class Exercise(object):
    def __init__(self, question, choices, correctAnswer, img=None, \
            explanation=None, googleMapPoints=None):
        self.question = question
        self.choices = choices
        # zjistime cislo spravne odpovedi
        for i, c in enumerate(choices):
            if c == correctAnswer:
                self.correctAnswer = i
                break
        else:
            raise Exception('correctAnswer is not among choices')
        self.image = img
        self.googleMapPoints = googleMapPoints

# TODO do vlastniho souboru
class ExerciseEvaluator(object):
    """ Evaluates relevance and probability of well formed of exercises.
    """
    DEFAULT_RANK = 0.3
    DEFAULT_RANK_LOW = 0.05
    DISCOUNT_FACTOR = 0.9
    RANDOM_BOOST = 0.1

    def __init__(self, historyRate, geographyRate, peopleRate):
        self.rank = {}
        self.rank['history'] = historyRate
        self.rank['geography'] = geographyRate
        self.rank['people'] = peopleRate
        self.rank['example'] = self.DEFAULT_RANK
        self.rank['reason'] = self.DEFAULT_RANK_LOW
        self.rank['number'] = self.DEFAULT_RANK
        self.rank['category'] = self.DEFAULT_RANK
        self.rank['image'] = self.DEFAULT_RANK
        self.rank['gap'] = self.DEFAULT_RANK_LOW

    def getRank(self, exerciseType, choiceQuality=1.0):
        assert choiceQuality > 0
        rank = self.rank[exerciseType] * choiceQuality
        if choiceQuality > 0.5:
            self.rank[exerciseType] *= self.DISCOUNT_FACTOR
            rank += random.random() * self.RANDOM_BOOST
        return rank


MAX_QUESTIONS = 15
def createExercises(ontology, bigrams, maxQuestions=MAX_QUESTIONS):
    # get topic rates
    historyScoreStr = ontology.value(EXERCISES.topic, EXERCISES.history)
    historyScr = float(historyScoreStr)
    geographyScoreStr = ontology.value(EXERCISES.topic, EXERCISES.geography)
    geographyScr = float(geographyScoreStr)
    peopleScoreStr = ontology.value(EXERCISES.topic, EXERCISES.people)
    peopleScr = float(peopleScoreStr)
    exerciseEvaluator = ExerciseEvaluator(historyScr, geographyScr, peopleScr)

    questions = []
    # bigramy by asi meli spise byt soucasti objektu ontologie (dedicnost)
    questions.extend(createWhenQuestions(ontology, exerciseEvaluator))
    questions.extend(createWhoQuestions(ontology, exerciseEvaluator))
    questions.extend(createWhereQuestions(ontology, exerciseEvaluator))
    questions.extend(createHowManyQuestions(ontology, exerciseEvaluator))
    questions.extend(createExampleQuestions(ontology,exerciseEvaluator,bigrams))
    questions.extend(createReasonQuestions(ontology,exerciseEvaluator,bigrams))
    questions.extend(createCategoryQuestions(ontology, exerciseEvaluator))
    questions.extend(imageQuestions(ontology, exerciseEvaluator, bigrams))

    #questions.extend(abstractQuestions(ontology))
    if len(questions) < maxQuestions:
        questions.extend(createGapQuestions(ontology,exerciseEvaluator,bigrams))
    
    #return [q for (rank, q) in sorted(questions, reverse=True)]

    questions = sorted(questions, reverse=True)[:maxQuestions]
    random.shuffle(questions)
    return [q for (rank, q) in questions]
    #return [{'question':'otazka', 'choices':['a','b','c'], 'correctAnswer':2},
    #        {'question':'otazka2', 'choices':['d','e','f'], 'correctAnswer':1}]


def createWhenQuestions(ontology, exerciseEvaluator):
    # TODO prepsat pomoci prepared queries (vse)
    events = ontology.query(
        """ SELECT DISTINCT ?label ?date
            WHERE {
                ?e a exrc:fact .
                ?e rdfs:label ?label .
                ?e exrc:when ?date .
            }""")
    events = [(json.loads(e), json.loads(d)) for (e, d) in events]
    dates = findAll('Date', ontology)
    exercises = []
    for label, date in events:
        question = formWQuestion(label, 'kdy')
        correctAnswer = date[1]
        choices = {correctAnswer}
        for i in range(3):
            choices.add(random.choice(dates))
        while len(choices) < 4:
            choices.add(unicode(int(correctAnswer) + random.randint(-75,75)))
        choices = list(choices)
        # TODO randomizovani odpovedi presunout do tride Exercise
        random.shuffle(choices)
        exercise = Exercise(question, choices, correctAnswer)
        exercises.append((exerciseEvaluator.getRank('history'), exercise))
    return exercises


def createExampleQuestions(ontology, exerciseEvaluator, bigrams):
    # TODO prepsat pomoci prepared queries (vse)
    facts = ontology.query(
        """ SELECT DISTINCT ?label ?example
            WHERE {
                ?e a exrc:fact .
                ?e rdfs:label ?label .
                ?e exrc:example ?example .
            }""")
    facts = [(json.loads(f), json.loads(e)) for (f, e) in facts]
    exercises = []
    for label, example in facts:
        if isVerb(label[-1]):
            question = sentenceFromPhrase(label, end=' například:')
        else:
            question = sentenceFromPhrase(label, end='. Uveďte příklad.')
        correctAnswer = example[1]
        choices = {correctAnswer}
        keywords, quality = findSimilarKeywords(example, ontology, bigrams)
        while len(choices) < 4:
            choices.add(random.choice(keywords))
        choices = list(choices)
        random.shuffle(choices)
        exercise = Exercise(question, choices, correctAnswer)
        exercises.append((exerciseEvaluator.getRank('example',quality), exercise))
    return exercises


def createCategoryQuestions(ontology, exerciseEvaluator):
    categories = ontology.query(
        """ SELECT DISTINCT ?category ?kwLabel ?kwType
            WHERE {
                ?kw a exrc:keyword .
                ?kw rdfs:label ?kwLabel .
                ?kw exrc:category ?category .
                OPTIONAL { ?kw exrc:type ?kwType }
            }""")

    allKeywords = findAllKeywords(ontology)
    if len(allKeywords) < 4:
        return [] # not enough keywords
    allPlaces = findAll('Place', ontology)
    allPeople = findAll('Person', ontology)

    categoryGroups = {}
    categoryTypes = {}
    for c in categories:
        category, keyword = unicode(c[0]), unicode(c[1])
        categoryGroups.setdefault(category, [])
        categoryGroups[category].append(keyword)
        if c[2] and len(categoryGroups.get(category, [])) <= 3:
            kwType = unicode(c[2])
            categoryTypes.setdefault((category, kwType), 0)
            categoryTypes[(category, kwType)] += 1

    exercises = []
    for category, group in categoryGroups.items():
        if len(group) >= 3:
            choices = group[:3]
            if categoryTypes.get((category, 'Date'), 0) > 0:
                continue
            peopleCount = categoryTypes.get((category, 'Person'), 0)
            placeCount = categoryTypes.get((category, 'Place'), 0)
            if peopleCount == 3:
                question = 'Kdo nepatří mezi ostatní?'
            elif peopleCount >= 1:
                question = 'Kdo nebo co nepatří mezi ostatní?'
            else:
                question = 'Co nepatří mezi ostatní?'
            if peopleCount == 3:
                # max 10 trials to select person not in the group
                for i in range(10):
                    answer = random.choice(allPeople)
                    if answer not in group:
                        choices.append(answer)
                        quality = 1
                        break
            if placeCount == 3:
                for i in range(10):
                    answer = random.choice(allPlaces)
                    if answer not in group:
                        choices.append(answer)
                        quality = 1
                        break
            if len(choices) < 4:
                quality = 0.1
                for i in range(10):
                    answer = random.choice(allKeywords)
                    if answer not in group:
                        choices.append(answer)
                        break
                else:
                    answer = randomNameGenerator()
                    while answer in choices:
                        answer = randomNameGenerator()
                    choices.append(answer)
            
            random.shuffle(choices)
            exercise = Exercise(question, choices, answer)
            exercises.append((exerciseEvaluator.getRank('category', quality),exercise))
    return exercises


def createGapQuestions(ontology, exerciseEvaluator, bigrams):
    # TODO prepsat pomoci prepared queries (vse)
    facts = ontology.query(
        """ SELECT DISTINCT ?label ?keyword
            WHERE {
                ?e a exrc:fact .
                ?e rdfs:label ?label .
                ?e exrc:includesKeyword ?keyword .
            }""")
    facts = [(json.loads(f), json.loads(e)) for (f, e) in facts]
    exercises = []
    for label, keyword in facts:
        correctAnswer = keyword[0]
        if len(keyword[2]) < 8 or keyword[2][6] != 'c':
            case = '1'
        else:
            case = keyword[2][7]
        for token in label:
            if token[0] == correctAnswer:
                if token[3] == 'Date':
                    token[0] = 'v roce ...'
                    correctAnswer = token[1]
                else:
                    token[0] = '...'
        label[0][0] = label[0][0][0].upper() + label[0][0][1:] # capital letter
        question = sentenceFromPhrase(label, beginning='Doplňte větu: ')
        choices = {correctAnswer}
        keywords, quality = findSimilarKeywords(keyword, ontology, bigrams)
        if case != 1:
            quality = quality * 0.8 # TODO nahodne zvolene cislo!
        while len(choices) < 4:
            kw = random.choice(keywords)
            if case != '1':
                kw = declineNounPhrase(kw, case)
            choices.add(kw)
        choices = list(choices)
        random.shuffle(choices)
        exercise = Exercise(question, choices, correctAnswer) 
        exercises.append((exerciseEvaluator.getRank('gap',quality), exercise))
    return exercises


def createHowManyQuestions(ontology, exerciseEvaluator):
    # TODO prepsat pomoci prepared queries (vse)
    facts = ontology.query(
        """ SELECT DISTINCT ?label ?number
            WHERE {
                ?e a exrc:fact .
                ?e rdfs:label ?label .
                ?e exrc:howMany ?number .
            }""")
    facts = [(json.loads(f), json.loads(n)) for (f, n) in facts]
    exercises = []
    for label, number in facts:
        question = sentenceFromPhrase(label, end='?') 
        correctAnswer = number[0]
        choices = {correctAnswer}
        percentage = ('%' in question)
        while len(choices) < 4:
            # TODO vhodnejsi pravdepobnostni rozlozeni
            if percentage:
                choices.add(unicode(random.randint(0,100)))
            else:
                choices.add(unicode(random.randint(0, 3*int(correctAnswer))))
        choices = list(choices)
        random.shuffle(choices)
        exercise = Exercise(question, choices, correctAnswer)
        exercises.append((exerciseEvaluator.getRank('number'), exercise))
    return exercises


def createReasonQuestions(ontology, exerciseEvaluator, bigrams):
    # TODO prepsat pomoci prepared queries (vse)
    facts = ontology.query(
        """ SELECT DISTINCT ?fact ?reason
            WHERE {
                ?e a exrc:fact .
                ?e rdfs:label ?fact .
                ?e exrc:why ?reason .
            }""")
    facts = [(json.loads(f), sentenceFromPhrase(json.loads(r), \
            beginning='Protože ', end='.')) for (f, r) in facts]
    allReasons = [r for (q, r) in facts]
    exercises = []
    for fact, reason in facts:
        # TODO osetrit kontextovost, ale tak abychom neprisli o vsechny otazky
        #if not isContextFree(fact):
        #    continue
        question = formWQuestion(fact, questionWord='proč')
        choices = {reason}
        if len(allReasons) < 5:
            choices.update(allReasons)
        else:
            for i in range(5):
                choices.add(random.choice(allReasons))
        quality = float(len(choices)) / 4
        while len(choices) < 4:
            choices.add(phraseGenerator(bigrams, beginning='Protože', end='.'))
        choices = list(choices)
        random.shuffle(choices)
        exercise = Exercise(question, choices, reason) 
        exercises.append((exerciseEvaluator.getRank('reason',quality), exercise))
    return exercises


def createWhoQuestions(ontology, exerciseEvaluator):
    # TODO prepsat pomoci prepared queries (vse)
    facts = ontology.query(
        """ SELECT DISTINCT ?label ?person
            WHERE {
                ?e a exrc:fact .
                ?e rdfs:label ?label .
                ?e exrc:who ?person .
            }""")
    facts = [(json.loads(f), json.loads(p)) for (f, p) in facts]
    # TODO tohle nevypada moc dobre, udelat vlasti tridu pro ontologii
    # dedenim a pridat metodu findAll
    people = findAll('Person', ontology)
    exercises = []
    for label, person in facts:
        correctAnswer = person[1]
        case = person[2][7]
        if case == '3':
            questionWord = 'komu'
        elif case == '7':
            questionWord = 'kým'
        elif case in ['2', '4']:
            questionWord = 'koho'
        else:
            questionWord = 'kdo'
        question = formWQuestion(label, questionWord)
        choices = {correctAnswer}
        # TODO to vybirani inteligentneji
        for i in range(4):
            choices.add(random.choice(people))
        quality = float(len(choices)) / 4
        while len(choices) < 4:
            choices.add(randomNameGenerator())
        choices = list(choices)
        random.shuffle(choices)
        exercise = Exercise(question, choices, correctAnswer)
        exercises.append((exerciseEvaluator.getRank('people', quality), exercise))
    return exercises


def createWhereQuestions(ontology, exerciseEvaluator):
    # TODO prepsat pomoci prepared queries (vse)
    places = findAll('Place', ontology)
    exercises = []
    questionWords = [
            ('where', 'kde'),
            ('whereFrom', 'odkud'),
            ('whereTo', 'kam'),
            ('whichPlace', 'které místo')]
    for questionWordEn, questionWordCs in questionWords:
        facts = ontology.query(
            """ SELECT DISTINCT ?label ?place
                WHERE {
                    ?e a exrc:fact .
                    ?e rdfs:label ?label .
                    ?e exrc:%s ?place .
                }""" % questionWordEn)
        facts = [(json.loads(f), json.loads(p)) for (f, p) in facts]
        for label, place in facts:
            correctAnswer = place[1]
            question = formWQuestion(label, questionWordCs)
            choices = {correctAnswer}
            for i in range(4):
                choices.add(random.choice(places))
            quality = float(len(choices)) / 4
            while len(choices) < 4:
                choices.add(randomPlaceGenerator())
            choices = list(choices)
            random.shuffle(choices)
            exercise = Exercise(question, choices, correctAnswer) 
            exercises.append((exerciseEvaluator.getRank('geography',quality),exercise))
    return exercises


def abstractQuestions(ontology):
    # TODO prepsat pomoci prepared queries (vse)
    abstracts = ontology.query(
        """ SELECT DISTINCT ?label ?abstract
            WHERE {
                ?keyword a exrc:keyword .
                ?keyword rdfs:label ?label .
                ?keyword exrc:abstract ?abstract .
            }""")
    keywords = ontology.query(
        """ SELECT DISTINCT ?label
            WHERE {
                ?keyword a exrc:keyword .
                ?keyword rdfs:label ?label .
            }""")
    keywords = [k[0] for k in keywords]
    if len(keywords) < 4:
        return []
    exercises = []
    for label, abstract in abstracts:
        question = createAbstractQuestion(unicode(label), unicode(abstract))
        choices = {label}
        while len(choices) < 4:
            # TODO inteligentnejsi vyber distraktoru
            choices.add(random.choice(keywords))
        choices = list(choices)
        random.shuffle(choices)
        exercises.append(Exercise(question, choices, label))
    return exercises


def imageQuestions(ontology, exerciseEvaluator, bigrams):
    # TODO prepsat pomoci prepared queries (vse)
    images = ontology.query(
        """ SELECT DISTINCT ?label ?src
            WHERE {
                ?img a exrc:image .
                ?img rdfs:label ?label .
                ?img exrc:url ?src .
            }""")
    labels = [l for (l, s) in images]
    exercises = []
    for label, src in images:
        question = 'Co je na obrázku?'
        choices = {label}
        for i in range(4):
            choices.add(random.choice(labels))
        quality = float(len(choices)) / 4
        while len(choices) < 4:
            r = random.randint(1,10)
            if r == 1:
                choices.add(phraseGenerator(bigrams))
            elif r == 2:
                choices.add(randomNameGenerator())
            elif r == 3:
                choices.add(randomPlaceGenerator())
            else:
                choices.add(random.choice(labels))
                quality += 0.25
        choices = list(choices)
        random.shuffle(choices)
        exercise = Exercise(question, choices, label, img=src)
        exercises.append((exerciseEvaluator.getRank('image', quality), exercise))
    return exercises


def createAbstractQuestion(keyword, abstract):
    gapAbstract = abstract.replace(keyword, '?????')
    # TODO use difflib.SequenceMatcher to disclose words too similar to the
    # keyword
    # TODO okrajovat prilis dlouhe abstracty
    question = 'Poznejte podle popisu:\n\n%s' % gapAbstract
    return question


def findSimilarKeywords(keyword, ontology, bigrams):
    keywords = set()
    if keyword[3] != 'kw':
        keywords.update(findAll(keyword[3], ontology))
    if keyword[3] == 'Date':
        while len(keywords) < 4:
            keywords.add(unicode(int(keyword[1]) + random.randint(-75,75)))
        return list(keywords), 1
    categoryGroups = findCategoryGroups(keyword[1], ontology)
    keywords.update([kw for group in categoryGroups.values() for kw in group])
    goodKeywordsRatio = (1+float(len(keywords)))/4
    if len(keywords) < 4:
        for kw in findAllKeywords(ontology):
            if len(keywords) < 4:
                keywords.add(kw)
        while len(keywords) < 4:
            if keyword[3] == 'Person':
                keywords.add(randomNameGenerator())
            elif keyword[3] == 'Place':
                keywords.add(randomPlaceGenerator())
            else:
                keywords.add(phraseGenerator(bigrams))
    return list(keywords), goodKeywordsRatio

# TODO asi by nemelo byt v tomto souboru, spise vytvorit objekt pro ontologii
# a z tohohle udelat metodu ontologie
def findCategoryGroups(keywordLemma, ontology):
    categories = ontology.query(
        """ SELECT DISTINCT ?category
            WHERE {
                ?kw a exrc:keyword .
                ?kw rdfs:label "%s" .
                ?kw exrc:category ?category .
            }""" % keywordLemma)
    categoryGroups = {}
    for c in categories:
        category = c[0]
        keywords = ontology.query(
            """ SELECT DISTINCT ?keywordLabel
                WHERE {
                    ?kw a exrc:keyword .
                    ?kw rdfs:label ?keywordLabel .
                    ?kw exrc:category "%s" .
                }""" % category)
        categoryGroups[unicode(category)] = [unicode(kw[0]) for kw in keywords]
    return categoryGroups


# Exercises using facts obtained through translation

def createCouldMeetQuestions(ontology):
    
    def intervalsIntersects(int1, int2):
        return int1[1] >= int2[0] and int2[1] >= int1[0]

    facts = ontology.query(
        """ SELECT DISTINCT ?personName ?birthYear ?deathYear
            WHERE {
                ?person exrc:type "Person" ;
                    rdfs:label ?personName ;
                    exrc:birthYear ?birthYear ;
                    exrc:deathYear ?deathYear .
            }""")
    lives = {p: (b, d) for (p, b, d) in facts}
    exercises = []
    for i, p in enumerate(lives.keys()):
        for q in lives.keys()[i+1:]:
            couldMeet = intervalsIntersects(lives[p], lives[q])
            question = "Mohli se teoreticky potkat %s a %s?" % (p, q)
            choices = ["Ano", "Ne"]
            correctAnswer = "Ano" if couldMeet else "Ne"
            exercise = Exercise(question, choices, correctAnswer)
            exercises.append(exercise)
    return exercises
    

RANGE_LAT = 1.0
RANGE_LNG = 3.0
def createEventOnMapQuestions(ontology):
    facts = ontology.query(
        """ SELECT DISTINCT ?eventLabel ?lat ?lng
            WHERE {
                ?event exrc:type "http://schema.org/Event" ;
                    rdfs:label ?eventLabel ;
                    exrc:latitude ?lat ;
                    exrc:longitude ?lng .
            }""")
    exercises = []
    for event, lat, lng in facts:
        # TODO sofistikovanejsi vyber distraktoru (takhle bude spravna odpoved
        # nejcasteji nekde uprostred)
        # TODO sklonovani "probehnout"
        question = "Najděte na mapě, kde proběhla %s." % event
        choices = ['A', 'B', 'C', 'D']
        correctAnswer = random.choice(choices)
        points = []
        lat = float(lat)
        lng = float(lng)
        for choice in choices:
            if choice == correctAnswer:
                points.append((lat, lng))
            else:
                newLat = lat + (2*random.random()-1)*RANGE_LAT
                newLng = lng + (2*random.random()-1)*RANGE_LNG
                points.append((newLat, newLng))
        exercise = Exercise(question, choices, correctAnswer, \
                googleMapPoints=points)
        exercises.append(exercise)
    return exercises


def createExercisesFromTranslations(ontology):
    exercises = []
    exercises.extend(createCouldMeetQuestions(ontology))
    exercises.extend(createEventOnMapQuestions(ontology))
    return exercises
